import {
  useAuthentication,
  useAuthCache,
  useAppDispatch,
  useSessionCache,
  useQuerybuilderHash,
  useMLEnabledHash,
  useQuerybuilderSettings,
  readInSchemaFromBackend,
  useQuerybuilder,
  useVisualization,
  wsSchemaRequest,
  wsSchemaSubscription,
  wsSchemaStatsRequest,
  useML,
} from '@/lib/data-access';
import { Broker, wsQueryRequest, wsQuerySubscription, wsQueryTranslationSubscription } from '@/lib/data-access/broker';
import { addInfo } from '@/lib/data-access/store/configSlice';
import { setMLResult } from '@/lib/data-access/store/mlSlice';
import { setQueryText, setQuerybuilderNodes, setQueryState, setActiveQueryID } from '@/lib/data-access/store/querybuilderSlice';
import { useEffect } from 'react';
import {
  wsGetState,
  wsGetStates,
  wsUpdateState,
  nilUUID,
  wsGetStatesSubscription,
  wsDeleteStateSubscription,
  wsGetStateSubscription,
  wsTestSaveStateConnectionSubscription,
  wsStateGetPolicy,
  DatabaseStatus,
} from '../broker/wsState';
import {
  addSaveState,
  testedSaveState,
  selectSaveState,
  updateSaveStateList,
  updateSelectedSaveState,
  setFetchingSaveStates,
  setStateAuthorization,
} from '../store/sessionSlice';
import { URLParams, getParam } from './url';
import { VisState, setVisualizationState } from '../store/visualizationSlice';
import { isEqual } from 'lodash-es';
import { setSchemaAttributeInformation, setSchemaLoading } from '../store/schemaSlice';
import { addError } from '@/lib/data-access/store/configSlice';
import { unSelect } from '../store/interactionSlice';
import { wsReconnectSubscription, wsUserGetPolicy } from '../broker/wsUser';
import { authorized, setSessionID } from '../store/authSlice';
import { asyncSetNewGraphQueryResult, queryingBackend } from '../store/graphQueryResultSlice';
import {
  SaveState,
  SchemaGraphStats,
  allMLTypes,
  LinkPredictionInstance,
  GraphQueryTranslationResultMessage,
  wsReturnKey,
} from 'ts-common';
import { SemSubstrVisComponent } from '@/lib/vis/visualizations';

export const EventBus = (props: { onRunQuery: () => void; onAuthorized: () => void }) => {
  const { login } = useAuthentication();
  const auth = useAuthCache();
  const dispatch = useAppDispatch();
  const ml = useML();
  const session = useSessionCache();
  const queryHash = useQuerybuilderHash();
  const queryBuilder = useQuerybuilder();
  const mlHash = useMLEnabledHash();
  const visState = useVisualization();
  const queryBuilderSettings = useQuerybuilderSettings();

  async function loadSaveState(saveStateID: string | undefined, saveStates: Record<string, SaveState>) {
    if (saveStateID && saveStates && saveStateID in saveStates) {
      console.debug('Setting state fetched from database', saveStateID, saveStates);
      const state = saveStates[saveStateID];
      if (state) {
        if (state.queryStates && state.queryStates.openQueryArray.length > 0) {
          let activeQuery = state.queryStates.openQueryArray.find(q => q.id === state.queryStates.activeQueryId);
          if (!activeQuery) {
            activeQuery = state.queryStates.openQueryArray[0];
            dispatch(setActiveQueryID(activeQuery.id || -1));
          }
          if (activeQuery) {
            const queryBuilderState = {
              ...state.queryStates,
              openQueryArray: state.queryStates.openQueryArray.map(query => ({
                ...query,
                ignoreReactivity: false,
                queryTranslation: {} as GraphQueryTranslationResultMessage,
              })),
            };

            dispatch(setQueryState(queryBuilderState));
            dispatch(setQuerybuilderNodes(activeQuery));
          }
        } else {
          console.error('No queries found in state', state);
        }
        dispatch(
          setVisualizationState(
            Object.keys(state.visualizations).length !== 0 // should only occur in mock data
              ? (state.visualizations as VisState)
              : {
                  activeVisualizationIndex: -1,
                  openVisualizationArray: [],
                },
          ),
        );
      }
    }
  }

  useEffect(() => {
    const unsubs: (() => void)[] = [];

    unsubs.push(
      wsSchemaSubscription(data => {
        dispatch(setSchemaLoading(false));
        if (!data) {
          dispatch(addError('Failed to fetch schema'));
          return;
        } else {
          dispatch(readInSchemaFromBackend(data));
          dispatch(addInfo('Schema graph updated'));
        }
      }),
    );

    unsubs.push(
      wsQuerySubscription(data => {
        if (!data) {
          dispatch(addError('Failed to fetch graph query result'));
          return;
        }
        asyncSetNewGraphQueryResult(data, dispatch);
        dispatch(addInfo('Query Executed!'));
        dispatch(unSelect());
      }),
    );

    unsubs.push(
      wsTestSaveStateConnectionSubscription(data => {
        if (data?.status === 'success' && data.saveStateID) {
          dispatch(testedSaveState(data.saveStateID));
        } else {
          dispatch(addError('DB connection failed'));
        }
      }),
    );

    unsubs.push(
      wsReconnectSubscription(data => {
        if (data == null) return;

        dispatch(setSessionID(data));

        props.onAuthorized();

        // Process URL Params
        const paramSaveState = getParam(URLParams.saveState);
        if (paramSaveState && paramSaveState !== nilUUID) {
          wsGetState({ saveStateID: paramSaveState });
        }

        dispatch(setFetchingSaveStates(true));

        // check authorizations
        wsUserGetPolicy(null, res => {
          if (res) dispatch(authorized(res));
        });

        wsGetStates(null, data => {
          if (data) {
            dispatch(setFetchingSaveStates(false));
            return true;
          } else {
            dispatch(addError('Failed to fetch states'));
            return false;
          }
        });

        // Broker.instance().sendMessage({ //TODO!!
        //   sessionID: auth?.sessionID || '',
        //   key: 'broadcastState',
        //   body: { type: 'subscribe', status: '', value: {} },
        // });
      }),
    );

    Broker.instance().subscribe((data: SchemaGraphStats) => {
      dispatch(setSchemaAttributeInformation(data));
      dispatch(addInfo('Received attribute information'));
    }, wsReturnKey.schemaStatsResult);

    // Broker.instance().subscribe((data: QueryBuilderState) => dispatch(setQuerybuilderNodes(data)), 'query_builder_state');

    allMLTypes.forEach(mlType => {
      const id = Broker.instance().subscribe((data: LinkPredictionInstance[]) => {
        dispatch(setMLResult({ type: mlType, result: data }));
      }, mlType);
      unsubs.push(() => Broker.instance().unSubscribe(mlType, id));
    });

    unsubs.push(
      wsGetStatesSubscription((data, status) => {
        console.debug('Save States updated', data, status);
        if (status !== 'success' || !data) {
          dispatch(addError('Failed to fetch states'));
          return;
        }

        dispatch(updateSaveStateList(data));
        const d = Object.fromEntries(data.map(x => [x.id, x]));
        loadSaveState(session.currentSaveState, d);
        data.forEach(ss => {
          if (session.saveStatesAuthorization?.[ss.id] === undefined) {
            wsStateGetPolicy({ saveStateID: ss.id }, ret => {
              if (!ret) {
                dispatch(addError('Failed to fetch state authorization'));
                return;
              }
              dispatch(setStateAuthorization({ id: ss.id, authorization: ret }));
            });
          }
        });
      }),
    );

    unsubs.push(
      wsGetStateSubscription((data, status) => {
        if (status !== 'success' || !data) {
          dispatch(addError('Failed to fetch state'));
          return;
        }
        if (data.id !== nilUUID) {
          console.info('Save State updated', data, status);
          dispatch(addSaveState(data));
          dispatch(selectSaveState(data.id));
          loadSaveState(data.id, session.saveStates);
          console.log('Querying', data.id);
          wsStateGetPolicy({ saveStateID: data.id }, ret => {
            if (!ret) {
              dispatch(addError('Failed to fetch state authorization'));
              return;
            }
            dispatch(setStateAuthorization({ id: data.id, authorization: ret }));
          });
        }
      }),
    );

    unsubs.push(wsDeleteStateSubscription(data => {}));

    // Broker.instance().subscribe((response: TestDatabaseConnectionResponse) => {
    //   if (response && response.status === 'success') dispatch(testedSaveState(response.saveStateID));
    // }, 'tested_connection');

    unsubs.push(
      wsQueryTranslationSubscription(response => {
        if (response && response.result) {
          dispatch(setQueryText(response));
        }
      }),
    );

    Broker.instance().subscribe((response: GraphQueryTranslationResultMessage) => {
      if (response && response.result) {
        dispatch(setQueryText(response));
      }
    }, wsReturnKey.queryStatusTranslationResult);

    login();

    // Setup cleanup
    return () => {
      // clear callback subscriptions
      unsubs.forEach(unsub => {
        unsub();
      });
      Broker.instance().unSubscribeAll(wsReturnKey.queryStatusTranslationResult);
      Broker.instance().unSubscribeAll(wsReturnKey.queryStatusResult);
      Broker.instance().unSubscribeAll(wsReturnKey.schemaInference);
    };
  }, []);

  useEffect(() => {
    if (session.currentSaveState) {
      const state = { ...session.saveStates[session.currentSaveState] };
      const openQueryArray = queryBuilder.openQueryArray.map(query => ({
        id: query.id,
        name: query.name,
        graph: query.graph,
        settings: query.settings,
        attributesBeingShown: query.attributesBeingShown,
      }));

      console.warn('Query State', openQueryArray);

      if (
        (state.queryStates == null ||
          state.queryStates.activeQueryId != queryBuilder.activeQueryId ||
          !isEqual(state.queryStates.openQueryArray, openQueryArray)) &&
        queryBuilder.openQueryArray.length > 0
      ) {
        state.queryStates = {
          activeQueryId: queryBuilder.activeQueryId,
          openQueryArray: openQueryArray,
        };
        dispatch(queryingBackend());
        dispatch(updateSelectedSaveState(state));
        console.log('Updating state', state);
        wsUpdateState(state, (data, status) => {
          if (status === 'unchanged') return;
          console.warn('Query State Updated', status);
          if (data) {
            console.warn('Query State Updated', data.queryStates.openQueryArray, openQueryArray, state);
            if (!isEqual(data.queryStates.openQueryArray, openQueryArray)) {
              dispatch(
                setQueryState({
                  activeQueryId: data.queryStates.activeQueryId,
                  openQueryArray: data.queryStates.openQueryArray.map(query => ({
                    ...query,
                    // TODO: remove this when the backend is fixed
                    ignoreReactivity: false,
                    queryTranslation: {} as GraphQueryTranslationResultMessage,
                  })),
                }),
              );
            }
          }
        });
      }
    }
  }, [queryBuilderSettings, queryBuilder.openQueryArray.map(x => x.name), queryHash]);

  useEffect(() => {
    if (session.currentSaveState) {
      const state = { ...session.saveStates[session.currentSaveState] };
      if (!isEqual(state.visualizations, visState)) {
        console.debug('Updating visState state', visState);
        state.visualizations = { ...visState };
        dispatch(updateSelectedSaveState(state));
        wsUpdateState(state);
      }
    }
  }, [visState]);

  useEffect(() => {
    // New active database
    if (session.currentSaveState && session.currentSaveState !== nilUUID) {
      dispatch(setSchemaLoading(true));
      wsSchemaRequest({ saveStateID: session.currentSaveState });
      wsSchemaStatsRequest({ saveStateID: session.currentSaveState });
      loadSaveState(session.currentSaveState, session.saveStates);
    }
  }, [session.currentSaveState]);

  useEffect(() => {
    // New active database
    if (session.currentSaveState && session.testedSaveState[session.currentSaveState] === DatabaseStatus.tested) {
      wsQueryRequest({ saveStateID: session.currentSaveState, ml });
    }
  }, [session.currentSaveState, session.testedSaveState]);

  useEffect(() => {
    // Newly (un)authorized
    if (auth.authentication?.authenticated && auth.authentication?.jwt) {
      console.warn('Authorized', auth.authentication);
      Broker.instance()
        .useAuth(auth.authentication)
        .connect(() => {
          console.debug('WS connected', session.currentSaveState, window.location.search);
        });
    } else {
      // dispatch(logout());
    }
  }, [auth.authentication]);

  return <div className="hide"></div>;
};
