import { createSlice, PayloadAction } from '@reduxjs/toolkit';
import type { RootState } from './store';
import { MultiGraph } from 'graphology';
import { QueryMultiGraphology as QueryGraphology } from '../../querybuilder/model/graphology/utils';
import { isEqual } from 'lodash-es';
import {
  Graph,
  Query,
  QueryBuilderSettings,
  GraphQueryTranslationResultMessage,
  QueryGraphEdgeHandle,
  QueryMultiGraph,
  QueryUnionType,
} from 'ts-common';

const defaultGraph: () => Graph = () => ({ nodes: [], edges: [], attributes: {}, options: {} });

export type QueryBuilderAttributeBeingShown = object;

export type QueryBuilderObject = Query & {
  ignoreReactivity: boolean;
  queryTranslation: GraphQueryTranslationResultMessage;
};

export type QueryBuilderState = {
  activeQueryId: number;
  openQueryArray: QueryBuilderObject[];
};

export type SchemaState = {
  id?: number;
  settings: Record<string, any>;
};

// Define the initial state using that type
export const initialState: QueryBuilderState = {
  activeQueryId: -1,
  openQueryArray: [
    {
      graph: defaultGraph(),
      ignoreReactivity: false,
      settings: {
        limit: 500,
        depth: { min: 1, max: 1 },
        layout: 'manual',
        autocompleteRelation: true,
        unionTypes: {},
      },
      queryTranslation: {
        queryID: '',
        result: '',
      },
      attributesBeingShown: [],
      name: 'Query',
    },
  ],

  // schemaLayout: 'Graphology_noverlap',
} as QueryBuilderState;

// The initialState but only with keys relevant to sending to the backend
export const initialStateBackend: QueryBuilderState = {
  activeQueryId: initialState.activeQueryId,
  openQueryArray: [
    (({ graph, settings, attributesBeingShown, name }) => ({ graph, settings, attributesBeingShown, name }))(
      initialState.openQueryArray[0],
    ),
  ],
} as QueryBuilderState;

export const querybuilderSlice = createSlice({
  name: 'querybuilder',
  initialState,
  reducers: {
    setQuerybuilderGraph: (state: QueryBuilderState, action: PayloadAction<QueryMultiGraph>) => {
      const activeQuery = getActiveQuery(state);

      if (activeQuery) {
        activeQuery.graph = action.payload;
        activeQuery.ignoreReactivity = false;
      }
    },
    /**
     * Sets the querybuilder nodes, settings, and attributes being shown,
     * if the payload contains the required information.
     * @param {QueryBuilderState} action.payload the payload with the new state
     */
    setQuerybuilderNodes: (state: QueryBuilderState, action: PayloadAction<Query>) => {
      const activeQuery = getActiveQuery(state);

      if (activeQuery && action.payload.graph?.nodes && action.payload.graph?.edges) {
        activeQuery.graph = action.payload.graph;
        activeQuery.settings = action.payload.settings;
        activeQuery.attributesBeingShown = action.payload.attributesBeingShown || [];
      }
      console.warn('111', state);
    },
    clearQB: (state: QueryBuilderState) => {
      state.activeQueryId = -1;
      state.openQueryArray = JSON.parse(JSON.stringify(initialState.openQueryArray));
    },
    setQuerybuilderSettings: (state: QueryBuilderState, action: PayloadAction<QueryBuilderSettings>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        activeQuery.settings = action.payload;
      }
    },
    setQueryText: (state: QueryBuilderState, action: PayloadAction<GraphQueryTranslationResultMessage>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        activeQuery.queryTranslation = action.payload;
      }
    },
    attributeShownToggle: (state: QueryBuilderState, action: PayloadAction<QueryGraphEdgeHandle>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        const existing = activeQuery.attributesBeingShown.findIndex(a => isEqual(a, action.payload));
        if (existing === -1) {
          activeQuery.attributesBeingShown.push(action.payload);
        } else {
          activeQuery.attributesBeingShown.splice(existing, 1);
        }
      }
    },
    setQueryUnionType: (state: QueryBuilderState, action: PayloadAction<{ nodeId: string; unionType: QueryUnionType }>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        if (activeQuery.settings.unionTypes == null) {
          activeQuery.settings.unionTypes = {};
        }
        activeQuery.settings.unionTypes[action.payload.nodeId] = action.payload.unionType;
      }
    },
    addNewQuery: (state: QueryBuilderState) => {
      const newQuery: QueryBuilderObject = JSON.parse(JSON.stringify(initialState.openQueryArray[0]));
      newQuery.name = `Query`;

      state.openQueryArray.push(newQuery);
      state.activeQueryId = -1;
    },
    removeQueryByID: (state: QueryBuilderState, action: PayloadAction<number>) => {
      state.openQueryArray = state.openQueryArray.filter(query => query.id !== action.payload);
      if (state.activeQueryId === action.payload) {
        const availableQuery = state.openQueryArray.find(query => query.id !== action.payload);
        state.activeQueryId = availableQuery && availableQuery.id ? availableQuery.id : -1;
      }
    },
    setActiveQueryID: (state: QueryBuilderState, action: PayloadAction<number>) => {
      if (state.openQueryArray.find(query => query.id === action.payload) === undefined) {
        const availableQuery = state.openQueryArray.find(query => query.id !== action.payload);
        state.activeQueryId = availableQuery && availableQuery.id ? availableQuery.id : -1;
        console.error('QueryID not found', action.payload, state.openQueryArray);
        return;
      }
      state.activeQueryId = action.payload;
    },
    setQueryName: (state: QueryBuilderState, action: PayloadAction<string>) => {
      const activeQuery = getActiveQuery(state);
      if (activeQuery) {
        activeQuery.name = action.payload;
      }
    },
    setQueryState: (state: QueryBuilderState, action: PayloadAction<QueryBuilderState>) => {
      state.openQueryArray = action.payload.openQueryArray;
      state.activeQueryId = action.payload.activeQueryId;
      const activeQuery = state.openQueryArray.find(q => q.id === state.activeQueryId);
      if (!activeQuery) {
        const availableQuery = state.openQueryArray?.[0];
        state.activeQueryId = availableQuery && availableQuery.id ? availableQuery.id : -1;
        console.error('QueryID not found');
      }
    },
  },
});

function getActiveQuery(state: QueryBuilderState) {
  let activeQuery = state.openQueryArray.find(q => q.id === state.activeQueryId);

  if (activeQuery == null) {
    activeQuery = state.openQueryArray[0];
  }

  return activeQuery;
}

export const queryBuilderState = (state: RootState): QueryBuilderState => state.querybuilder;
export const queryBuilderSettingsState = (state: RootState): QueryBuilderSettings => {
  const activeQuery = getActiveQuery(state.querybuilder);
  if (!activeQuery) {
    return { limit: 500, depth: { min: 1, max: 1 }, layout: 'manual', autocompleteRelation: true, unionTypes: {} };
  }
  return activeQuery?.settings;
};

export const setQuerybuilderGraphology = (payload: QueryGraphology) => {
  return querybuilderSlice.actions.setQuerybuilderGraph(payload.export());
};

/** Select the querybuilder nodes in serialized fromat */
export const toQuerybuilderGraphology = (graph: QueryMultiGraph): QueryGraphology => {
  const ret = new QueryGraphology();
  ret.import(MultiGraph.from(graph).export());
  return ret;
};

/** Select the querybuilder nodes and convert it to a graphology object */
export const selectQuerybuilderGraph = (state: RootState): QueryMultiGraph => {
  // This is really weird but for some reason all the attributes appeared as read-only otherwise
  const activeQuery = getActiveQuery(state.querybuilder);
  return activeQuery?.graph as QueryMultiGraph;
};

/** Select the querybuilder nodes and convert it to a graphology object */
export const selectQuerybuilderHash = (state: RootState): string => {
  const activeQuery = getActiveQuery(state.querybuilder);
  if (!activeQuery?.graph) {
    console.error('No active graph found');
    return ''; // Return an empty string if no active graph exists
  }
  const hashedNodes = activeQuery.graph.nodes.map(n => {
    const node = { ...n };
    if (n?.attributes) {
      const newAttributes = { ...n?.attributes };
      newAttributes.x = 0;
      newAttributes.y = 0;
      newAttributes.height = 0;
      newAttributes.width = 0;
      node.attributes = newAttributes;
    }
    return node;
  });
  return JSON.stringify({
    nodes: hashedNodes,
    edges: activeQuery.graph.edges.map(n => ({
      key: n.key,
      source: n.source,
      target: n.target,
    })),
    attributesBeingShown: activeQuery.attributesBeingShown,
  });
};

// /**
//  * selects the SchemaLayout enum
//  * @param {GraphLayout} state
//  * @returns {GraphLayout} enum of type GraphLayout
//  */
// export const selectSchemaLayout = (state: RootState) =>
//   state.schema.schemaLayout;

export default querybuilderSlice.reducer;
export const {
  setQuerybuilderGraph,
  clearQB,
  setQuerybuilderSettings,
  setQuerybuilderNodes,
  setQueryText,
  attributeShownToggle,
  setQueryUnionType,
  setQueryName,
  addNewQuery,
  removeQueryByID,
  setActiveQueryID,
  setQueryState,
} = querybuilderSlice.actions;

export const queryBuilderAttributesShown = (state: RootState) => {
  const activeQuery = getActiveQuery(state.querybuilder);
  return activeQuery?.attributesBeingShown || [];
};
