import { useMemo, useState } from 'react';
import { ControlContainer, TooltipProvider, Tooltip, TooltipTrigger, Button, TooltipContent, Input } from '../../components';
import { Popover, PopoverTrigger, PopoverContent } from '../../components/layout/Popover';
import {
  useActiveSaveStateAuthorization,
  useAppDispatch,
  useGraphQueryResult,
  useML,
  useQuerybuilderSettings,
  useQuerybuilder,
} from '../../data-access';
import {
  clearQB,
  setQuerybuilderSettings,
  addNewQuery,
  removeQueryByID,
  setActiveQueryID,
  setQueryName,
} from '../../data-access/store/querybuilderSlice';
import { QueryMLDialog } from './querysidepanel/QueryMLDialog';
import { QuerySettings } from './querysidepanel/QuerySettings';
import { ManualQueryDialog } from './ManualQueryDialog';
import { wsManualQueryRequest } from '../../data-access/broker';
import { Tabs, Tab } from '../../components/tabs';

export type QueryBuilderToggleSettings = 'settings' | 'ml' | 'logic' | 'relatedNodes' | undefined;

export type QueryBuilderNavProps = {
  toggleSettings: QueryBuilderToggleSettings;
  onFitView: () => void;
  onRunQuery: () => void;
  onScreenshot: () => void;
  onLogic: () => void;
};

export const QueryBuilderNav = (props: QueryBuilderNavProps) => {
  const dispatch = useAppDispatch();
  const { activeQueryId, openQueryArray } = useQuerybuilder();
  const qb = useQuerybuilderSettings();
  const result = useGraphQueryResult();
  const resultSize = useMemo(() => (result ? result.edges.length : 0), [result]);
  const ml = useML();
  const saveStateAuthorization = useActiveSaveStateAuthorization();
  const [editingIdx, setEditingIdx] = useState<{ idx: number; text: string } | null>(null);

  /**
   * Clears all nodes in the graph.
   */
  function clearAllNodes() {
    dispatch(clearQB());
  }

  const mlEnabled = ml.linkPrediction.enabled || ml.centrality.enabled || ml.communityDetection.enabled || ml.shortestPath.enabled;

  const handleManualQuery = (query: string) => {
    wsManualQueryRequest({ query: query });
  };

  return (
    <div className="sticky shrink-0 top-0 flex items-stretch justify-between h-7 bg-secondary-100 border-b border-secondary-200 max-w-full">
      <div className="flex items-center">
        <h1 className="text-xs font-semibold text-secondary-600 px-2 truncate">Query builder</h1>
        <TooltipProvider delayDuration={0}>
          <Tooltip>
            <TooltipTrigger>
              <Button
                as="a"
                variantType="secondary"
                variant="ghost"
                size="xs"
                iconComponent="icon-[ic--baseline-add]"
                onClick={async () => {
                  dispatch(addNewQuery());
                }}
              />
            </TooltipTrigger>
            <TooltipContent>
              <p>Add query</p>
            </TooltipContent>
          </Tooltip>
        </TooltipProvider>
      </div>
      <Tabs>
        {openQueryArray
          .filter(query => query.id != null)
          .map((query, i) => (
            <Tab
              text=""
              activeTab={query.id === activeQueryId}
              key={i}
              onClick={() => {
                if (query.id == null) return;
                dispatch(setActiveQueryID(query.id));
              }}
            >
              <>
                {editingIdx?.idx === i ? (
                  <Input
                    type="text"
                    size="xs"
                    value={editingIdx.text}
                    label=""
                    onChange={e => {
                      setEditingIdx({ idx: i, text: e });
                    }}
                    onBlur={() => {
                      dispatch(setQueryName(editingIdx.text));
                      setEditingIdx(null);
                    }}
                    onKeyDown={e => {
                      if (e.key === 'Enter') {
                        dispatch(setQueryName(editingIdx.text));
                        setEditingIdx(null);
                      }
                    }}
                    className="w-20"
                  />
                ) : (
                  <div
                    onDoubleClick={e => {
                      e.stopPropagation();
                      dispatch(setActiveQueryID(query.id || -1));
                      setEditingIdx({ idx: i, text: query.name ?? '' });
                    }}
                  >
                    {query.name ?? 'Query'}
                  </div>
                )}
                {openQueryArray.filter(query => query.id != null).length > 1 && (
                  <Button
                    variantType="secondary"
                    variant="ghost"
                    disabled={!saveStateAuthorization.database?.W}
                    rounded
                    size="2xs"
                    iconComponent="icon-[ic--baseline-close]"
                    onClick={e => {
                      e.stopPropagation();
                      if (query.id !== undefined) {
                        dispatch(removeQueryByID(query.id));
                      }
                    }}
                  />
                )}
              </>
            </Tab>
          ))}
      </Tabs>
      <div className="sticky right-0 px-0.5 ml-auto items-center flex truncate">
        <ControlContainer>
          <TooltipProvider delayDuration={0}>
            <Tooltip>
              <TooltipTrigger>
                <Button
                  variantType="secondary"
                  variant="ghost"
                  size="xs"
                  iconComponent="icon-[ic--baseline-fullscreen]"
                  onClick={props.onFitView}
                />
              </TooltipTrigger>
              <TooltipContent>
                <p>Fit to screen</p>
              </TooltipContent>
            </Tooltip>
            <Tooltip>
              <TooltipTrigger>
                <Button
                  variantType="secondary"
                  variant="ghost"
                  size="xs"
                  disabled={!saveStateAuthorization.query.W}
                  iconComponent="icon-[ic--baseline-delete]"
                  onClick={() => {
                    if (saveStateAuthorization.query.W) clearAllNodes();
                  }}
                />
              </TooltipTrigger>
              <TooltipContent>
                <p>Clear query panel</p>
              </TooltipContent>
            </Tooltip>
            <Tooltip>
              <TooltipTrigger>
                <Button
                  variantType="secondary"
                  variant="ghost"
                  size="xs"
                  iconComponent="icon-[ic--baseline-camera-alt]"
                  onClick={props.onScreenshot}
                />
              </TooltipTrigger>
              <TooltipContent>
                <p>Capture screen</p>
              </TooltipContent>
            </Tooltip>
            <Popover>
              <PopoverTrigger>
                <Tooltip>
                  <TooltipTrigger>
                    <Button
                      variantType="secondary"
                      variant="ghost"
                      size="xs"
                      disabled={!saveStateAuthorization.query.W}
                      iconComponent="icon-[ic--baseline-settings]"
                      className="query-settings"
                    />
                  </TooltipTrigger>
                  <TooltipContent>
                    <p>Query builder settings</p>
                  </TooltipContent>
                </Tooltip>
              </PopoverTrigger>
              <PopoverContent>
                <QuerySettings />
              </PopoverContent>
            </Popover>
            <Tooltip>
              <TooltipTrigger>
                <Button
                  variantType="secondary"
                  variant="ghost"
                  size="xs"
                  iconComponent="icon-[ic--baseline-cached]"
                  onClick={props.onRunQuery}
                />
              </TooltipTrigger>
              <TooltipContent>
                <p>Rerun query</p>
              </TooltipContent>
            </Tooltip>
            <Tooltip>
              <TooltipTrigger asChild>
                <Button
                  variantType="secondary"
                  variant="ghost"
                  size="xs"
                  disabled={!saveStateAuthorization.query.W}
                  iconComponent="icon-[ic--baseline-difference]"
                  onClick={props.onLogic}
                />
              </TooltipTrigger>
              <TooltipContent disabled={props.toggleSettings === 'logic'}>
                <p>Logic settings</p>
              </TooltipContent>
            </Tooltip>
            <Popover>
              <PopoverTrigger>
                <Tooltip>
                  <TooltipTrigger>
                    <Button
                      variantType={mlEnabled ? 'primary' : 'secondary'}
                      variant="ghost"
                      size="xs"
                      disabled={!saveStateAuthorization.query.W}
                      iconComponent="icon-[ic--baseline-lightbulb]"
                      className={mlEnabled ? 'border-primary-600' : ''}
                    />
                  </TooltipTrigger>
                  <TooltipContent disabled={props.toggleSettings === 'ml'}>
                    {mlEnabled ? (
                      <>
                        <p className="font-bold text-base">Machine learning</p>
                        <p className="mb-2">Algorithms detected the following results:</p>

                        {ml.linkPrediction.enabled && ml.linkPrediction.result && (
                          <>
                            <p className="mt-2 font-semibold">Link prediction</p>
                            <p>{ml.linkPrediction.result.length} links</p>{' '}
                          </>
                        )}
                        {ml.centrality.enabled && Object.values(ml.centrality.result).length > 0 && (
                          <>
                            <p className="mt-2 font-semibold">Centrality</p>
                            <p>
                              {Object.values(ml.centrality.result)
                                .reduce((a, b) => b + a)
                                .toFixed(2)}{' '}
                              sum of centers
                            </p>
                          </>
                        )}
                        {ml.communityDetection.enabled && ml.communityDetection.result && (
                          <>
                            <p className="mt-2 font-semibold">Community detection</p>
                            <p>{ml.communityDetection.result.length} communities</p>
                          </>
                        )}
                        {ml.shortestPath.enabled && (
                          <>
                            <p className="mt-2 font-semibold">Shortest path</p>
                            {ml.shortestPath.result?.length > 0 && <p># of hops: {ml.shortestPath.result.length}</p>}
                            {!ml.shortestPath.srcNode ? (
                              <p>Please select source node</p>
                            ) : (
                              !ml.shortestPath.trtNode && <p>Please select target node</p>
                            )}
                          </>
                        )}
                      </>
                    ) : (
                      <p>Machine learning</p>
                    )}
                  </TooltipContent>
                </Tooltip>
              </PopoverTrigger>
              <PopoverContent>
                <QueryMLDialog />
              </PopoverContent>
            </Popover>
            <Popover>
              <PopoverTrigger>
                <Tooltip>
                  <TooltipTrigger>
                    <Button
                      variantType={qb.limit <= resultSize ? 'primary' : 'secondary'}
                      variant="ghost"
                      size="xs"
                      disabled={!saveStateAuthorization.query.W}
                      iconComponent="icon-[ic--baseline-filter-alt]"
                      className={qb.limit <= resultSize ? 'border-primary-600' : ''}
                    />
                  </TooltipTrigger>
                  <TooltipContent disabled={props.toggleSettings === 'ml'}>
                    <p className="font-bold text-base">Limit</p>
                    <p>Limits the number of edges retrieved from the database.</p>
                    <p>Required to manage performance.</p>
                    <p className={`font-semibold${qb.limit <= resultSize ? ' text-primary-800' : ''}`}>
                      Fetched {resultSize} of a maximum of {qb.limit} edges
                    </p>
                  </TooltipContent>
                </Tooltip>
              </PopoverTrigger>
              <PopoverContent>
                <div className="flex flex-col w-full gap-2 px-4 py-2">
                  <span className="text-xs font-bold">Limit</span>
                  <Input
                    type="number"
                    size="xs"
                    value={qb.limit}
                    lazy
                    label=""
                    onChange={e => {
                      dispatch(setQuerybuilderSettings({ ...qb, limit: Number(e) }));
                    }}
                    className={`w-24${qb.limit <= resultSize ? ' border-danger-600' : ''}`}
                    containerClassName=""
                  />
                </div>
              </PopoverContent>
            </Popover>
            <Popover>
              <PopoverTrigger disabled={!saveStateAuthorization.query.W}>
                <Tooltip>
                  <TooltipTrigger>
                    <Button variantType="secondary" variant="ghost" size="xs" iconComponent="icon-[ic--baseline-search]" />
                  </TooltipTrigger>
                  <TooltipContent>
                    <p>Manual Query</p>
                  </TooltipContent>
                </Tooltip>
              </PopoverTrigger>
              <PopoverContent>
                <ManualQueryDialog onSubmit={handleManualQuery} />
              </PopoverContent>
            </Popover>
          </TooltipProvider>
        </ControlContainer>
      </div>
    </div>
  );
};
